##
## DataDome f5 BigIP iRule module.
##
## Version 1.6.1
## DataDome (https://datadome.co)
##

when RULE_INIT {
    # License key (DATADOME_LICENSE_KEY)
    set static::DATADOME_Key "PUT YOUR KEY HERE!"

    # If you would like to inject datadome JS tag, just uncomment code below and put valid JS license key
    # set static::DATADOME_js_target {</head>}
    # set static::DATADOME_js {
    #     <script>
    #     !function(a,b,c,d,e,f){a.ddjskey=e;a.ddoptions=f||null;var m=b.createElement(c),n=b.getElementsByTagName(c)[0];m.async=1,m.src=d,n.parentNode.insertBefore(m,n)}(window,document,"script","https://js.datadome.co/tags.js","YOUR_DATADOME_JS_KEY");
    #     </script>
    # }

    # API Connection virtual server
    set static::DATADOME_vs /Common/datadome_api_ssl_vs

    # API connection timeout in milliseconds
    set static::DATADOME_timeout 150

    # URIRegex (DATADOME_URI_REGEX) and URIRegexExclusion (DATADOME_URI_REGEX_EXCLUSION)
    # are regex used to match URI.
    #
    # The logic is:
    #   Does URI match with URIRegexExclusion if present?
    #     if yes stop
    #       if no Does URI match with URIRegex if present?
    #       if no stop
    #       if yes, send to API
    #
    # Example with a URIRegexExclusion:
    #   static::DATADOME_URI_REGEX_EXCLUSION /\.(avi|flv|mka|mkv|mov|mp4|mpeg|mpg|mp3|flac|ogg|ogm|opus|wav|webm|webp|bmp|gif|ico|jpeg|jpg|png|svg|svgz|swf|eot|otf|ttf|woff|woff2|css|less|js)$/
    #
    # Default behavior excludes static assets.
    # To disable a regex, unset it, do not just disable it.
    if {[info exists static::DATADOME_URI_REGEX]} {
        unset static::DATADOME_URI_REGEX
    }
    set static::DATADOME_URI_REGEX_EXCLUSION {(?i)\.(avi|flv|mka|mkv|mov|mp4|mpeg|mpg|mp3|flac|ogg|ogm|opus|wav|webm|webp|bmp|gif|ico|jpeg|jpg|png|svg|svgz|swf|eot|otf|ttf|woff|woff2|css|less|js)$}

    # An option to easy log headers that datadome inject into request or response
    set static::DATADOME_LOG_REQUEST_HEADERS 0
    set static::DATADOME_LOG_RESPONSE_HEADERS 0

    # Logging request performance
    set static::DATADOME_LOG_REQUEST_TIME 0

    # HSL publisher
    # You should create a log publisher in System->Logs->Configuration->Log Publishers.
    # If this line is commented out, this module may flood the local log if you enabled logging for each request
    #set static::DATADOME_LOG_PUBLISHER /Common/lpDataDome
    # or you can use UDP pool
    #set static::DATADOME_LOG_POOL syslog_server_pool

    # init http client
    call http_req_rule_init
}

when CLIENT_ACCEPTED {
    set protocol "http"
    if {[info exists static::DATADOME_LOG_PUBLISHER]} {
        set hsl [HSL::open -publisher $static::DATADOME_LOG_PUBLISHER]
    } elseif {[info exists static::DATADOME_LOG_POOL]} {
        set hsl [HSL::open -proto UDP -pool $static::DATADOME_LOG_POOL]
    }
}

# remove it if virtual server doesn't support HTTPS
when CLIENTSSL_HANDSHAKE {
    set protocol "https"
    if {[info exists static::DATADOME_LOG_PUBLISHER]} {
        set hsl [HSL::open -publisher $static::DATADOME_LOG_PUBLISHER]
    } elseif {[info exists static::DATADOME_LOG_POOL]} {
        set hsl [HSL::open -proto UDP -pool $static::DATADOME_LOG_POOL]
    }
}

when HTTP_REQUEST {
    set uri [HTTP::uri]
    set path [HTTP::path]

    set datadome_matched 0

    set TIME_start [clock clicks -milliseconds]

    if {[info exists static::DATADOME_URI_REGEX_EXCLUSION] && [regexp $static::DATADOME_URI_REGEX_EXCLUSION $path]} {
        return
    }

    if {[info exists static::DATADOME_URI_REGEX] && ![regexp $static::DATADOME_URI_REGEX $path]} {
        return
    }

    set datadome_matched 1

    set xForwardedFor [HTTP::header "X-Forwarded-For"]
    set xForwardedForLength [string length $xForwardedFor]
    set xForwardedForLimit 512

    if { $xForwardedForLength > $xForwardedForLimit } {
        set xForwardedFor [string range $xForwardedFor [expr $xForwardedForLength - $xForwardedForLimit] $xForwardedForLength]
    }

    # Session By Header
    set clientId [string range [HTTP::cookie value "datadome"] 0 128]
    set xDatadomeClientId [HTTP::header "X-DataDome-ClientID"]
    set datadomeHeaders [list \
        Content-Type "application/x-www-form-urlencoded" \
        User-Agent "DataDome" \
    ]

    if { [string length $xDatadomeClientId] != 0 } {
        set clientId $xDatadomeClientId
        lappend datadomeHeaders X-DataDome-X-Set-Cookie "true"
    }

    set body [list \
                  Key               $static::DATADOME_Key \
                  RequestModuleName "iRule" \
                  ModuleVersion     "1.7.0" \
                  ServerName        [string range [LB::server name] 0 512] \
                  IP                [getfield [IP::client_addr] "%" 1] \
                  Port              [TCP::client_port] \
                  TimeRequest       [concat [clock clicks -milliseconds]000] \
                  Protocol          [concat $protocol] \
                  Method            [HTTP::method] \
                  ServerHostname    [string range [HTTP::host] 0 512] \
                  Request           [string range [concat $uri] 0 2048] \
                  HeadersList       [string range [join [HTTP::header names] ","] 0 512] \
                  Host              [string range [HTTP::host] 0 512] \
                  UserAgent         [string range [HTTP::header "User-Agent"] 0 768] \
                  Referer           [string range [HTTP::header "Referer"] 0 1024] \
                  Accept            [string range [HTTP::header "Accept"] 0 512] \
                  AcceptEncoding    [string range [HTTP::header "Accept-Encoding"] 0 128] \
                  AcceptLanguage    [string range [HTTP::header "Accept-Language"] 0 256] \
                  AcceptCharset     [string range [HTTP::header "Accept-Charset"] 0 128] \
                  Origin            [string range [HTTP::header "Origin"] 0 512] \
                  XForwardedForIP   $xForwardedFor \
                  X-Requested-With  [string range [HTTP::header "X-Requested-With"] 0 128] \
                  Connection        [string range [HTTP::header "X-cnection"] 0 128] \
                  Pragma            [string range [HTTP::header "Pragma"] 0 128] \
                  From              [string range [HTTP::header "From"] 0 128] \
                  CacheControl      [string range [HTTP::header "Cache-Control"] 0 128] \
                  ContentType       [string range [HTTP::header "Content-Type"] 0 128] \
                  CookiesLen        [string length [HTTP::header "Cookie"]] \
                  AuthorizationLen  [string length [HTTP::header "Authorization"]] \
                  PostParamLen      [HTTP::header "Content-length"] \
                  ClientID          $clientId \
                  X-Real-IP         [string range [HTTP::header "X-Real-IP"] 0 128] \
                  Via               [string range [HTTP::header "Via"] 0 256] \
                  TrueClientIP      [string range [HTTP::header "True-Client-IP"] 0 128] \
                  SecCHUAMobile     [string range [HTTP::header "Sec-CH-UA-Mobile"] 0 8] \
                  SecFetchUser      [string range [HTTP::header "Sec-Fetch-User"] 0 8] \
                  SecChDeviceMemory [string range [HTTP::header "Sec-CH-Device-Memory"] 0 8] \
                  SecCHUAArch       [string range [HTTP::header "Sec-CH-UA-Arch"] 0 16] \
                  SecCHUAPlatform   [string range [HTTP::header "Sec-CH-UA-Platform"] 0 32] \
                  SecFetchDest      [string range [HTTP::header "Sec-Fetch-Dest"] 0 32] \
                  SecFetchMode      [string range [HTTP::header "Sec-Fetch-Mode"] 0 32] \
                  SecFetchSite      [string range [HTTP::header "Sec-Fetch-Site"] 0 64] \
                  SecCHUA           [string range [HTTP::header "Sec-CH-UA"] 0 128] \
                  SecCHUAModel      [string range [HTTP::header "Sec-CH-UA-Model"] 0 128] \
                  SecCHUAFullVersionList [string range [HTTP::header "Sec-CH-UA-Full-Version-List"] 0 256] \
                 ]

    set rheaders {}

    set TIME_taken_before_request [expr [clock clicks -milliseconds] - $TIME_start]

    # hostname here doesn't affect used backend, update VS to change it!
    set sts [call http_req \
                 -virt $static::DATADOME_vs -timeout $static::DATADOME_timeout \
                 -method POST -hostname "api.datadome.co" -path "/validate-request/" \
                 -headers $datadomeHeaders \
                 -form $body -redir 0 \
                 -rheaders rheaders -rbody rbody
            ]

    set TIME_taken_after_request [expr [clock clicks -milliseconds] - $TIME_start - $TIME_taken_before_request]

    set has_correct_x_datadome_response 0

    foreach {nam val} $rheaders {
        if {[string tolower $nam] eq "x-datadomeresponse"} {
            if {$sts eq $val} {
                set has_correct_x_datadome_response 1
            }
            break
        }
    }

    if {$has_correct_x_datadome_response ne "1"} {
        set rheaders {}
        return
    }

    call execute_datadome_header "X-DataDome-Request-Headers" $rheaders $static::DATADOME_LOG_REQUEST_HEADERS "request"

    set TIME_taken [expr [clock clicks -milliseconds] - $TIME_start]

    if {$static::DATADOME_LOG_REQUEST_TIME} {
        call datadome_info_log "DataDome spent: ${TIME_taken} ms in total, ${TIME_taken_before_request} ms to create request, and ${TIME_taken_after_request} to execute it."
    }

    if {$sts == 403 || $sts == 401 || $sts == 301 || $sts == 302} {
        eval [concat [list HTTP::respond $sts -version 1.1 content "$rbody" noserver] $rheaders]
        return
    }
}

when HTTP_RESPONSE {
    STREAM::disable

    if {$datadome_matched eq "1" && [HTTP::header value "Content-Type"] contains "text/html" && [info exists static::DATADOME_js] && [info exists static::DATADOME_js_target]} {
        STREAM::expression "@$static::DATADOME_js_target@$static::DATADOME_js$static::DATADOME_js_target@"
        STREAM::enable
    }
}

when HTTP_RESPONSE_RELEASE {
    if {$datadome_matched eq "1"} {
        call execute_datadome_header "X-DataDome-headers" $rheaders $static::DATADOME_LOG_RESPONSE_HEADERS "response"
    }
}

proc datadome_info_log {message} {
    if {[info exists hsl]} {
        # 134 is similar to local0.info
        HSL::send $hsl "<134> ${message}"
    } else {
        log local0.info "${message}"
    }
}

proc datadome_err_log {message} {
    if {[info exists hsl]} {
        # 131 is similar to local0.err
        HSL::send $hsl "<131> ${message}"
    } else {
        log local0.err "${message}"
    }
}

proc execute_datadome_header {name headers debug target} {
    foreach {nam val} $headers {
        if {[string tolower $nam] eq [string tolower $name]} {
            set candidates [regexp -inline -all -- {\S+} $val]
            foreach {cn cv} $headers {
                if {[lsearch -exact $candidates $cn] >= 0} {
                    if {$debug} {
                        call datadome_info_log "DataDome ${target} headers: ${cn}: ${cv}"
                    }
                    if {[string tolower $cn] eq "set-cookie"} {
                        HTTP::header insert $cn $cv
                    } else {
                        HTTP::header replace $cn $cv
                    }
                }
            }
            break
        }
    }
}

# Code below based on Mark's HTTP Super SIDEBAND Requestor! (v2.1)

proc http_req_rule_init {} {
 #IPv6 hostname per rfc2732; we recognize IPv6 but don't validate strictly
 set static::DataDome_urlsplit {(?i)^(https?)://(?:([^:@]+):?([^@]*)@)?((?:\x5b[0-9a-f:.]{2,45}(?:%[0-9a-z:.]{1,18})?\x5d)|(?:[0-9.a-z-]{1,255}))(?::([0-9]{0,5}))?([*]|(?:/[^\x23]*)?)(?:\x23.*|)$}

 set static::DataDome_ipv6match {(?i)[\x5b]?([0-9a-f:.]{2,45}(?:%[0-9a-z:.]{1,18})?)[\x5d]?}

 set static::DataDome_ckymatch {^([^=\x24\x5c\x7b\x7d\x20\t]+)="?([^\x22;,\x20\t\x5c]*)"?(?:[\x20\t]*;[\x20\t]*(.*))?$}

 set static::DataDome_quiver {(?i)<meta[\x20\t][^>]*[\x20\t]?http-equiv=['\x22]?content-type['\x22\x20\t]?[^>]*>}

 set static::DataDome_csetmeta {(?i)<meta[\x20\t]+charset=['\x22]?([^'\x22>]+)}

 set static::DataDome_xmlenc {(?i)<[?]xml[\x20\t]+.*encoding=['\x22]?([^'\x22?>]+)}

 #string option names (sort alphabetically)
 set static::DataDome_s_opts [list body cookies form headers hostname key \
                     method nocache ns passwd path proxy rbody rcookies \
                     rformat rheaders state tag type uri userid virt]
 #integer options (sort these too)
 set static::DataDome_n_opts [list debug expect idle maxbody munch \
                     redir sniff timeout]
 #and these as well...
 set static::DataDome_done [list accept accept-charset accept-encoding \
                   accept-language content-length content-type host \
       user-agent x-hssr-helper]
 set static::DataDome_nogo [list closed closing failed timeout]
}

proc http_req {args} {
 #set me [lindex [info level 0] 0]
 #oops, in TMOS 11.4-5 [info level 0] unavailing and [info frame] undefined
 set me "http_req"

 set uri ""
 set virt "!"
 set schm ""
 set ruid ""
 set rpwd ""
 set hostname ""
 set host ""
 set port ""
 set path ""
 set state ""
 set tag "HSSR"
 set key ""
 set rbody ""
 set rformat ""
 set rheaders ""
 set rcookies ""
 set method "GET"
 set form [list]
 set body ""
 set type ""
 set headers [list "!" "!"]
 set cookies [list]
 set userid "\x19"
 set passwd ""
 set proxy "!"
 set ns "!"

 set expect 0
 set nocache false
 set sniff 2
 set redir 1
 set munch 1
 set timeout 7500
 set idle 300
 set maxbody 1048576
 set debug 0

 set e ""
 set visits 0
 while {[incr visits] < 2} {
  if {[llength $args] & 1} {
   set e "arguments must be '-option value' pairs"
  } else {
   foreach {n v} $args {
    set opt [string range $n 1 end]
    set xn -1
    if {!($n starts_with "-") ||
        ( ([lsearch -sorted $static::DataDome_s_opts $opt] < 0) &&
          ([set xn [lsearch -sorted $static::DataDome_n_opts $opt]] < 0) )} {
     set e "option '${n}' not recognized"
     break
    }
    if {($xn >= 0) && [regexp {[^0-9]} $v]} {
     set e "option ${n} value must be integer"
     break
    }
    set $opt $v
   }
  }

  #evade iRules checker so we can be called from events
  #on different kinds of virtual server, etc
  if {($key eq "") &&
      ![catch {eval "IP::client_addr"} key] &&
      ![catch {eval "IP::protocol"} pn] &&
      ([set qry [expr {($pn == 6) ? "TCP::client_port" :
                        (($pn == 17) ? "UDP::client_port" : "")}]] ne "") &&
      ![catch {eval $qry} tmp]} {
   append key ":" $tmp
  }
  set tkm "${tag} ${key}: ${me}"

  if {$e ne ""} { break }

  #saved state (possibly reuse connection)?
  if {$state ne ""} { upvar 1 $state stvar }
  if {![info exists stvar] || ([llength $stvar] < 9)} {
   set stvar [list false {} "" "" "" "" [list] "" ""]
  } elseif {$debug && [lindex $stvar 0]} {
   call datadome_info_log "${tkm} might re-use conn to [lindex $stvar 2]"
  }
  foreach {connected conn cdest vs px nsvr crumbs creds kopf} $stvar { break }

  if {$virt eq "!"} { set virt $vs }
  if {$proxy eq "!"} { set proxy $px }
  if {$ns ne "!"} { set nsvr [expr {($ns ne "") ? "@${ns}" : ""}] }

  set ffcount [llength $form]
  set bodylen [string bytelength $body]

  set auth_hdrs ""

  if {$uri ne "" && ![regexp $static::DataDome_urlsplit $uri junk schm ruid rpwd host port path] ||
            ([string length $path] > 2040)} {
   set e "malformed -uri ${uri}"
  } elseif {([set schm [string tolower $schm]] ne "") &&
            [set rTLS [expr {[string index $schm end] eq "s"}]] &&
            ($virt eq "")} {
   set e "SIDEBAND to TLS (https) needs helper virtual-server"
  } elseif {$ffcount & 0x1} {
   set e "-form value must list pairs of field names and values"
  } elseif {$ffcount && ($bodylen || ($type ne ""))} {
   set e "option -form precludes options -body or -type"
  } elseif {$bodylen && (($method eq "GET") || ($method eq "HEAD"))} {
   set e "option -body incompatible with -method ${method}"
  } elseif {($path eq "*") && ($method ne "OPTIONS")} {
   set e "URI path '*' valid only with -method OPTIONS"
  } elseif {$ruid ne ""} {
   set e "username+password in request URI not supported"
  } elseif {[llength $headers] & 1} {
   set e "-headers option invalid"
  } elseif {[llength $cookies] & 1} {
   set e "-cookies option invalid"
  } elseif {$expect && ( ($expect != 100) || !($bodylen || $ffcount) ||
                         ($method eq "GET") || ($method eq "HEAD") )} {
   set e "use '-expect 100' only when sending -body or -form data"
  } elseif {$proxy ne ""} {
   if {![regexp $static::DataDome_urlsplit $proxy junk pxy_s pxy_uid pxy_pwd pxy_h pxy_p] ||
       ([string length $pxy_p] > 2040)} {
    set e "-proxy URL invalid"
    break
   }
   set pxy_s [string tolower $pxy_s]
   set pxy_h [string tolower $pxy_h]
   if {[string index $pxy_h 0] eq "\x5b"} {
    set pxy_h [string map [list "%25" "%"] $pxy_h]
    regexp $static::DataDome_ipv6match $pxy_h junk raddr
   } else {
    if {[string index $pxy_h end] eq "."} {
     set pxy_h [string range $pxy_h 0 end-1]
    }
    if {![catch {IP::addr $pxy_h mask 255.255.255.255}]} {
     set raddr [eval format "::ffff:%02x%02x:%02x%02x" [split $pxy_h "."]]
    } else {
     if {!( ([set tmp [lindex [eval "RESOLV::lookup ${nsvr} inet -a ${pxy_h}"] 0]] ne "") &&
            ([set raddr [eval format "::ffff:%02x%02x:%02x%02x" [split $tmp "."]]] ne "")
          ) &&
         ([set raddr [lindex [eval "RESOLV::lookup ${nsvr} inet6 -aaaa ${pxy_h}"] 0]] eq "") &&
         ($virt eq "")} {
      set e "cannot resolve proxy host ${pxy_h} to IP address"
      break
     }
    }
   }
   if {$pxy_p eq ""} {
    set pxy_p [expr {([string index $pxy_s end] eq "s") ? "443" : "80"}]
   }
   set xhelper ""
   if {$virt eq ""} {
    set dest "${raddr}.${pxy_p}"
   } else {
    set dest $virt
    if {$raddr ne ""} {
     set xhelper "X-DataDome-SSL-Helper: [list $pxy_s $raddr $pxy_p [virtual] $key]\r\n"
    }
   }
   if {$pxy_uid ne ""} {
    append auth_hdrs "Proxy-Authorization: Basic\x20" \
                     [b64encode "${pxy_uid}:${pxy_pwd}"] "\r\n"
   }
   if {$debug} {
    call datadome_info_log "${tkm} will send requests via proxy ${pxy_h}:${pxy_p}"
   }
  } else {
   set xhelper [set dest ""]
  }
 } ; #one-visit block
 if {$e ne ""} {
  call datadome_err_log "${tkm} ${e}"
  return -code error "${me} ${e}"
 }

 set host [string tolower $host]
 if {[set ipv6 [expr {[string index $host 0] eq "\x5b"}]]} {
  set host [string map [list "%25" "%"] $host]
  regexp $static::DataDome_ipv6match $host junk haddr
 } elseif {[string index $host end] eq "."} {
  set host [string range $host 0 end-1]
 }
 if {$port eq "" && $schm ne ""} { set port [expr {$rTLS ? "443" : "80"}] }
 if {$port eq ""} { set port "0" }
 set server "${host}:${port}"

 if {$path eq ""} { set path "/" }

 #milk helps us swallow cookies
 set milk [expr {$munch && ![regexp {[^.][.][^.]+[^0-9\x5d]$} $host]}]

 if {$userid eq ""} {
  set creds ""
 } elseif {$userid ne "\x19"} {
  set creds "Authorization: Basic [b64encode "${userid}:${passwd}"]\r\n"
 }
 if {$creds ne ""} { append auth_hdrs $creds }

 #reuse headers like User-Agent from saved state unless caller overrides
 if {($kopf eq "") || ![llength $headers] ||
     ([lindex $headers 0] ne "!") || ([lindex $headers 1] ne "!")} {
  set kopf ""

  set v [expr {([set x [lsearch -regexp $headers {(?i)^Accept$}]] < 0) ? \
               "*/*" : [lindex $headers [incr x]]}]
  append kopf "Accept: ${v}\r\n"

  set v [expr {([set x [lsearch -regexp $headers {(?i)^Accept-Charset$}]] < 0) ? \
               "utf-8;q=1, iso-8859-1;q=0.8, *;q=0.6" : \
               [lindex $headers [incr x]]}]
  append kopf "Accept-Charset: ${v}\r\n"

  set v [expr {([set x [lsearch -regexp $headers {(?i)^Accept-Encoding$}]] < 0) ? \
               "identity" : [lindex $headers [incr x]]}]
  append kopf "Accept-Encoding: ${v}\r\n"

  set v [expr {([set x [lsearch -regexp $headers {(?i)^Accept-Language$}]] < 0) ? \
               "en" : [lindex $headers [incr x]]}]
  append kopf "Accept-Language: ${v}\r\n"

  upvar #0 tcl_platform static::tcl_platform ; #SOL14544 workaround
  set tmm_vers $static::tcl_platform(tmmVersion)
  set v [expr {([set x [lsearch -regexp $headers {(?i)^User-Agent$}]] < 0) ? \
               "BigIP/${tmm_vers} (HSSRv2.1)" : [lindex $headers [incr x]]}]
  append kopf "User-Agent: ${v}\r\n"

  if {[llength $headers] > 1} {
   foreach {n v} $headers {
    if {($n ne "!") &&
        ([lsearch -sorted $static::DataDome_done [string tolower $n]] < 0)} {
     append kopf "${n}:\x20${v}\r\n"
    }
   }
  }
 }

 #$crumbs format {name1 {dom1 path1 pathlen1 val1 exp1 secure1} name2 ...}
 if {[llength $cookies]} {
  set cdom [expr {$milk ? "***" : \
                  [expr {([string first "." $host] > 0) ? \
                         "*.${host}" : $host}] }]
  set cpth "/"
  set cpsz 1
  set cexp 2147485546 ; #we yield one second to propitiate Saturn
  set cscr 0

  foreach {n v} $cookies {
   set repl 0
   #cookie names and paths ARE case-sensitive
   foreach x [lsearch -all -exact $crumbs $n] {
    set c [lindex $crumbs [incr x]]
    if {([lindex $c 0] eq $cdom) && ([lindex $c 1] eq $cpth)} {
     set crumbs [lreplace $crumbs $x $x [lreplace $c 3 5 $v $cexp $cscr]]
     set repl 1
    }
   }
   if {!$repl} {
    lappend crumbs $n [list $cdom $cpth $cpsz $v $cexp $cscr]
   }
  }
 }

 if {$ffcount} {
  #prepare to submit form
  set w3c [list "%20" "+" "%0a" "%0d%0a"]
  set utf8 0
  set payload ""
  set sep ""
  foreach {n v} $form {
   append payload $sep [string map $w3c [URI::encode $n]]
   append payload "=" [string map $w3c [URI::encode $v]]
   incr utf8 [expr {![string is ascii "${n}${v}"]}]
   set sep "&"
  }

  if {($method eq "GET") || ($method eq "HEAD")} {
   append path [expr {([string first "?" $path] < 0) ? "?" : "&"}]
   append path $payload
  } else {
   set body $payload
   set bodylen [string bytelength $body]

   set type "application/x-www-form-urlencoded"
   #we only assert charset= if payload contains non-ASCII so server
   #will (generally) assume an acceptable charset, since we don't
   #really know what server wants and ASCII is broadly compatible
   if {$utf8} { append type "; charset=UTF-8" }
  }
 }

 set maxclock [expr {[clock clicks -milliseconds] + $timeout}]

 #max redirects we will follow
 set rtry 7

 set rslt 555
 set hlist [list]
 set booty "" ; set bootylen 0
 set mimetype "unk/unk"
 set charset "unk"
 set who [set where [set how [set what [set when ""]]]]
 while {[incr rtry -1] >= 0} {
  #to whom will we connect?
  if {$dest eq ""} {
   if {$ipv6} {
    set raddr $haddr
   } elseif {![catch {IP::addr $host mask 255.255.255.255}]} {
    set raddr [eval format "::ffff:%02x%02x:%02x%02x" [split $host "."]]
   } else {
    if {!( ([set tmp [lindex [eval "RESOLV::lookup ${nsvr} inet -a ${host}"] 0]] ne "") &&
           ([set raddr [eval format "::ffff:%02x%02x:%02x%02x" [split $tmp "."]]] ne "")
        ) &&
        ([set raddr [lindex [eval "RESOLV::lookup ${nsvr} inet6 -aaaa ${host}"] 0]] eq "") &&
        ($virt eq "")} {
     set e "cannot resolve ${host} to IP address"
     set rtry -2
     break
    }
   }
   set xhelper ""
   if {$virt eq ""} {
    set dest "${raddr}.${port}"
   } else {
    set dest $virt
    if {$raddr ne ""} {
     set xhelper "X-DataDome-SSL-Helper: [list $schm $raddr $port [virtual] $key]\r\n"
    }
   }
  }

  #canonicalize path
  regsub -all {(?://)|(?:[^/]+/[.][.]/)} $path "/" path

  #retry POST/whatever after redirect(s)?
  if {($what ne "") && ($who eq $server) && ($path eq $where)} {
   set method $how
   set body $what
   set what ""
   set expect $when
   set when ""
  }

  #cons up HTTP request
  set req [expr {($proxy eq "") ?
                 "${method} ${path} HTTP/1.1\r\n" :
                 "${method} ${schm}://${server}${path} HTTP/1.1\r\n"
                }]
  append req $xhelper
  append req $auth_hdrs

  if {$hostname eq ""} { set hostname $server }
  append req "Host: ${hostname}\r\n"

  if {($method ne "GET") && ($method ne "HEAD")} {
   #we handle Content-Length because caller might confuse chars and bytes
   append req "Content-Length: ${bodylen}\r\n"

   if {$bodylen} {
    if {$type ne ""} { append req "Content-Type: ${type}\r\n" }

    set dyt [clock format [clock clicks -milliseconds] -format "%a, %d %b %Y %T GMT" -gmt 1]
    append req "Date:\x20${dyt}\r\n"

    if {$expect} { append req "Expect: 100-Continue\r\n" }
   }
  }

  if {$nocache} { append req "Cache-Control: no-cache\r\nPragma: no-cache\r\n" }

  #purge expired cookies
  set tmp [list]
  set now [clock clicks -milliseconds]
  foreach {n c} $crumbs {
   if {[lindex $c 4] > $now} { lappend tmp $n $c }
  }
  set crumbs $tmp

  #assemble Cookie header
  if {[llength $crumbs]} {
   set tmp ""
   set plen [string length $path]
   foreach {n c} $crumbs {
    set cdom [lindex $c 0]
    set cpth [lindex $c 1]
    if {($rTLS || ![lindex $c 5]) && (
         ($ipv6 && ([string index $cdom 0] eq "\x5b") &&
          [IP::addr [string range $cdom 1 end-1] equals $haddr]) ||
         [string match $cdom $host] ||
         ( ([string index $cdom 0] eq "*") &&
           ([string range $cdom 2 end] eq $host)
         )
        ) && (
         ($path starts_with $cpth) &&
         ( ($plen == [set cpsz [lindex $c 2]]) ||
           ([string index $path $cpsz] eq "/") ||
           ([string index $cpth end] eq "/")
         )
        )} {
     append tmp "\x20${n}=[lindex $c 3];"
    }
   }
   #possibly jar holds no cookies for this host
   if {$tmp ne ""} { append req "Cookie:${tmp}\r\n" }
  }

  #bring in previously-prepared standard and custom headers
  append req $kopf

  append req "\r\n" ; #headers complete

  #Are we already or still connected?
  if {$connected &&
      ( ($cdest ne $dest) ||
        [catch {connect info -status $conn} sts] ||
        ([lsearch -sorted $static::DataDome_nogo [lindex [lindex $sts 0] 0]] > -1)
      )} {
   catch {close $conn}
   set connected false
  }
  #otherwise a little race...

  if {!$connected} {
   if {$debug} {
    call datadome_info_log "${tkm} attempt connect to ${dest} '${server}' ${timeout} msec timeout"
   }
   if {[catch {connect -timeout $timeout -idle $idle -status sts $dest} conn] ||
       ($conn eq "")} {
    set e "connect to ${dest} '${server}' fails: ${sts} (${conn})"
    set rtry -2
    break
   }
   set connected true
   set cdest $dest
  }

  if {$debug} {
   call datadome_info_log "${tkm} send to ${dest} '${server}' ${bodylen}-octet body, headers=${req}"
  }

  #we accept only one 100-Continue per POST/PUT attempt
  set pending 0
  set bites 2 ; #at the apple, lover
  while {[incr bites -1] >= 0} {
   if {$expect} {
    if {!$pending} {
     #I see you shiver with antici...
     set pending $bodylen ; set bodylen 0
    } else {
     #...pation
     set bodylen $pending ; set pending 0
    }
   }

   if {$req ne ""} {
    if {$debug} {
     call datadome_info_log "${tkm} connected to ${dest} '${server}' sending [string length $req]-octet req headers"
     if {$pending} {
      call datadome_info_log "${tkm} expecting 100-Continue before sending body"
     }
    }
    if {[catch {send -timeout $timeout -status sts $conn $req} sent] ||
        ($sts ne "sent") || ($sent != [string length $req])} {
     set e "send [string length $req]-octet req to ${dest} '${server}' fails: sent ${sent}, got status ${sts} (${sent})"
     break
    }
   }
   if {$bodylen} {
    if {$debug} {
     call datadome_info_log "${tkm} sending ${bodylen}-octet body to ${dest} '${server}'"
    }
    if {[catch {send -timeout $timeout -status sts $conn $body} sent] ||
        ($sts ne "sent") || ($sent != $bodylen)} {
     set e "send $bodylen-octet body to ${dest} '${server}' fails: sent ${sent}, got status ${sts} (${sent})"
     break
    }
   }

   set ctr 50
   set intvl [expr {int($timeout / $ctr)}] ; #keep this with initial ctr value
   set hdrs [set booty ""]
   set sts "unk"
   while {($ctr > 0) && ($sts ne "failed") && ($sts ne "closed")} {
    if {[catch {recv -peek -timeout $intvl -status sts $conn} reply]} {
     set e "peek hdrs fails: ${sts} (${reply})"
     break
    }

    if {[set hbrk [string first "\r\n\r\n" $reply]] > 0} {
     #TODO: support rfc5987 header-value charsets (priority: quite low)
     if {[catch {recv -status sts [incr hbrk 4] $conn} hdrs]} {
      set e "recv hdrs fails: ${sts} (${hdrs})"
      set hdrs ""
      break
     }
     if {$debug} {
      call datadome_info_log "${tkm} reply hdrs [string map [list "\r\n" ",\x20"] $hdrs]"
     }

     #extract HTTP status code to $rslt
     if {([string length $hdrs] < 16) ||
         ![regexp {^HTTP/1[.][01]\x20([0-9]+)[^\r]*\r\n} $hdrs skip rslt]} {
      set e "server data unrecognized (not HTTP?), initial bytes:\x20"
      append e [URI::encode [string range $hdrs 0 31]]
      set hdrs ""
      break
     }
     set hdrs [string range $hdrs [string length $skip] end]

     if {($method eq "HEAD") || ($rslt == 204) || ($rslt starts_with "1")} {
      break
     }

     if {[regexp {(?i)Transfer-Encoding:[\x20\t]*Chunked} $hdrs]} {
      set bootylen 0
      set trail false
      while {$ctr > 0} {
       if {[catch {recv -peek -timeout $intvl -status sts $conn} chunk]} {
        set e "peek chunksz fails: ${sts} (${chunk})"
        break
       }

       if {$trail} {
        #flush optional trailing headers
        if {[string length $chunk] > 1} {
         if {( ($chunk starts_with "\r\n") && [set flush 2] ) ||
             ( ([set flush [string first "\r\n\r\n" $chunk]] >= 0) &&
               [incr flush 4] )} {
          if {[catch {recv -status sts $flush $conn} junk]} {
           set e "skip trailing headers fails: ${sts} (${junk})"
          }
          set trail false
          break
         }
        }
        incr ctr -1
        continue
       }

       set hex ""
       if {[regexp {^(?:[\r\n])*([0-9A-Fa-f]+)[^\r]*\r\n} $chunk skip hex] &&
           ($hex ne "0")} {
        if {[catch {recv -status sts [string length $skip] $conn} junk]} {
         set e "skip chunksz fails: ${sts} (${junk})"
         break
        }
        scan $hex "%x" chunksz
        if {$chunksz > 0} {
         if {($bootylen + $chunksz) > $maxbody} {
          set e "chunked body length exceeds -maxbody ${maxbody}"
          break
         }
         set z [expr {$ctr * $intvl}]
         if {![catch {recv -timeout $z -status sts $chunksz $conn} chunk] &&
             ([string length $chunk] == $chunksz)} {
          set booty [binary format a*a* $booty $chunk]
          incr bootylen $chunksz
         } else {
          set e "recv 0x${hex}=${chunksz}-octet chunk fails: ${sts} (${chunk})"
          break
         }
        }
       } elseif {$hex eq "0"} {
        if {[catch {recv -status sts [string length $skip] $conn} junk]} {
         set e "skip final chunksz fails: ${sts} (${junk})"
         break
        }
        set trail true ; #ignore optional trailing headers
        continue
       }
       incr ctr -1
      } ; #chunks loop
      if {$e eq ""} {
       if {$hex ne "0"} {
        set e "timed-out awaiting additional (or terminating zero-size) chunk"
       } elseif {$trail} {
        set e "timed-out while flushing optional trailing headers"
       }
      }
     } elseif {(![set clh [regexp {(?i)Content-Length:[\x20\t]*([0-9]+)} $hdrs junk clen]] &&
                [set clen $maxbody]) ||
               (($clen > 0) && ($clen <= $maxbody))} {
      set z [expr {$ctr * $intvl}]
      if {[catch {recv -timeout $z -status sts $clen $conn} booty] ||
          ([set bootylen [string length $booty]] < 0)} {
       set e "recv ${clen}-octet body fails: ${sts} (${booty})"
      } elseif {$clh && ($bootylen < $clen)} {
       set e "recv body fails (${sts}), got only ${bootylen} of Content-Length: ${clen} octets"
      }
     } elseif {$clen > $maxbody} {
      set e "Content-Length: ${clen} exceeds -maxbody ${maxbody}"
     }
     if {$e ne ""} {
      set booty "" ; set bootylen 0
     }
     #here we have reply headers, plus body if any, so stop polling
     break
    }

    incr ctr -1
   } ; #receive-poll (ctr) loop
   if {$e ne ""} {
    set rslt 555
    catch {close $conn}
    set connected false
    break
   }

   if {($hdrs eq "") && $expect && $pending && ($sts eq "timeout")} {
    if {$debug} {
     call datadome_info_log "${tkm} wait for 100-Continue timed out--per rfc2616 send data now"
    }
    set req ""
    continue
   }

   if {$debug} {
    if {$ctr < 1} {
     call datadome_info_log "${tkm} timed out awaiting response from ${dest} '${server}'"
    }
    call datadome_info_log "${tkm} status ${rslt}, got [string length $hdrs] headers octets and ${bootylen} content octets"
   }
   if {$rslt != 100} {
    set bites -1
   } elseif {!$expect} {
    #rfc2616 allows unexpected 100-Continue
    set req ""
    set bodylen 0
    if {$debug} {
     call datadome_info_log "${tkm} got unexpected 100-Continue from ${server}"
    }
   }

   if {([clock clicks -milliseconds] > $maxclock)} {
    if {$debug} {
     call datadome_info_log "${tkm} hit maxtime secs after [expr {7 - $rtry}] redirects"
    }
    set rtry -1
    break
   }

   #compute default cookie path
   regexp {^/[^?#]*} $path ckypath
   if {[set x [string last "/" $ckypath]] > 0} { incr x -1 }
   set ckypath [string range $ckypath 0 $x]

   #remove LWS from headers
   set hdrs [regsub -all {[\x20\t]*\r\n[\x20\t]+} $hdrs "\x20"]

   #parse headers
   set mimetype "unk/unk"
   set charset "unk"
   set hlist [list]

   set hdrs [string map [list "\r\n" "\n"] $hdrs]
   set rlist [split $hdrs "\n"]
   foreach rline $rlist {
    if {![regexp {^([^:]+):[ \t]+(.*)} $rline junk n v]} { continue }
    set v [string trim $v]

    #I wish (September 2013) iRules had "switch -nocase"
    switch -- [string tolower $n] {
     "set-cookie" {
      set cky [set val [set opts ""]]
      #unlike some browsers we reject nameless cookies
      while {([string bytelength $v] < 4480) &&
             [regexp $static::DataDome_ckymatch $v junk cky val opts]} {
       if {[regexp {(?i)Domain=[.]?([^.;\x20\t][^;\x20\t]*)} $opts junk cdom]} {
        set cdom [string tolower $cdom]
        if {[string index $cdom 0] eq "\x5b"} {
         if {!$ipv6 || ![regexp $static::DataDome_ipv6match $cdom junk caddr] ||
             ![IP::addr $caddr equals $haddr]} {
          if {!$milk} { break }
          set cdom "***"
         }
        } elseif {[string first "." $cdom] < 1} {
         if {$cdom ne $host} {
          if {!$milk} { break }
          set cdom "***"
         }
        } elseif {![string match [set cdom "*.${cdom}"] $host]} {
         if {!$milk} { break }
         set cdom "***"
        }
       } else {
        set cdom [expr {$milk ? "***" : $host}]
       }

       #rfc6265 has ?query in cookie-path but not MS-IE and not us
       if {![regexp {(?i)Path=(/[^?#;]*)} $opts junk cpth]} {
        set cpth $ckypath
       }
       regsub -all {(?://)|(?:[^/]+/[.][.]/)} $cpth "/" cpth
       set cpsz [string length $cpth]

       set cexp 2147485546
       if {[regexp {(?i)Max-Age=([0-9]+)} $opts junk expdt]} {
        set cexp [expr {$expdt + [clock clicks -milliseconds]}]
       } elseif {[regexp {(?i)Expires=([^;]+);?} $opts junk expdt] &&
                 ![catch {clock scan $expdt -gmt 1} expdt]} {
        set cexp $expdt
       }

       set cscr [regexp {(?i)\ySecure;?\y} $opts]

       #cookie names and paths ARE case-sensitive
       set repl 0
       foreach x [lsearch -all -exact $crumbs $cky] {
        set c [lindex $crumbs [incr x]]
        if {([lindex $c 0] eq $cdom) && ([lindex $c 1] eq $cpth)} {
         set crumbs [lreplace $crumbs $x $x [lreplace $c 3 5 $val $cexp $cscr]]
         set repl 1
        }
       }
       if {!$repl} {
        lappend crumbs $cky [list $cdom $cpth $cpsz $val $cexp $cscr]
       }
       #expired cookies will be purged before next request
       break
      }
     }
     "location" {
      if {$redir && ($rslt > 300) && (($rslt < 304) || ($rslt == 307))} {
       if {($rslt <= 302) && ($redir == 1) && ($what eq "") &&
           (($method ne "GET") && ($method ne "HEAD"))} {
        set who $server
        set where $path
        set how $method
        set what $body
        set when $expect
       }
       if {$rslt <= 303} {
        set method "GET"
        set body ""
        set expect 0
       }
       if {$v starts_with "/"} {
        #relative path so contact same host
        if {![regexp {^/[^\x23]*} $v a] || ([string length $a] > 2040)} {
         set e "bogus redirect from ${dest} '${server}'"
         set rtry -2
         break
        }
        set s $schm
        set r $rTLS
        set h $host
        set p $port
       } elseif {[regexp $static::DataDome_urlsplit $v junk s toss lose h p a] &&
                 ([string length $a] < 2041) && ($a ne "*")} {
        set s [string tolower $s]
        set r [expr {[string index $s end] eq "s"}]
        set h [string tolower $h]
        if {[string index $h 0] eq "\x5b"} {
         set h [string map [list "%25" "%"] $h]
        } elseif {[string index $h end] eq "."} {
         set h [string range $h 0 end-1]
        }
        if {$p eq ""} { set p [expr {$r ? "443" : "80"}] }
  if {$a eq ""} { set a "/" }
       } else {
        #oops, Location not HTTP URI or path bogus
        set e "cannot follow redir from ${dest} '${server}' to ${v}"
        set rtry -2
        break
       }
      }
      #don't molest host/port/etc before swallowing all cookies
     }
     "connection" {
      if {$connected && ($rslt != 100) && [regexp {(?i)\yclose\y} $v]} {
       catch {close $conn}
       set connected false
      }
     }
     "content-length" {
      if {($v != $bootylen) && ($method ne "HEAD")} {
       #rfc2616 sec4.4 commands us to log
       call datadome_err_log "${tkm} ${bootylen}-octet chunked body != Content-Length: ${v} from ${dest} '${server}'"
      }
     }
     "content-type" {
      set v [string tolower $v]
      regexp {^[^/;\x20\t]+/[^;\x20\t]+} $v mimetype
      if {![regexp {(?i)charset=([^;\x20\t]+)} $v junk charset] &&
          (($mimetype starts_with "text/") ||
           ($mimetype eq "application/javascript") ||
           ($mimetype eq "application/ecmascript")) && ($sniff != 2)} {
       set charset [expr {($sniff == 0) ? "iso-8859-1" : "utf-8"}]
      }
     }
     "www-authenticate" {
      if {($rslt == 401) && ($creds ne "") && [regexp {(?i)Basic\x20} $v]} {
       #We already sent "Authorization: Basic XXX" header but some servers
       #(f5 APM) will 401 first request and only look for creds on second
       #request (and change cookies too)
       set rslt 455
      }
     }
     "proxy-authenticate" {
      if {($rslt == 407) && ($pxy_uid ne "") && [regexp {(?i)Basic\x20} $v]} {
       set rslt 455 ; #what, again?
      }
     }
     "retry-after" {
      if {[regexp {^(503)|(30[1237])$} $rslt] && (
           ([regexp {^[0-9]+} $v delay] && $delay) ||
           ( ![catch {clock scan $v -gmt 1} dttm] &&
             ([set delay [expr {$dttm - [clock clicks -milliseconds]}]] > -1) )
          ) && (
           [set delay [expr {int(min($delay,($rtry * $timeout / 2)))}]] &&
           (($delay + [clock clicks -milliseconds] + $timeout) < $maxclock)
          )} {
       if {$debug} {
        call datadome_info_log "${tkm} Retry-After, sleep ${delay} seconds"
       }
       after $delay ; #sleep
       if {$rslt == 503} { set rslt 455 }
       #otherwise follow redirect
      }
     }
    } ; #header name switch

    lappend hlist $n $v
   } ; #header parsing loop
  } ; #100-continue loop
  if {$e ne ""} {
   set rtry -2
   break
  }

  if {$rslt == 455} {
   #retry Basic authz, or after server-requested delay
   continue
  }

  #do we follow a redirect, or are we finished?
  if {!$redir || ($rslt < 301) || (($rslt > 303) && ($rslt != 307))} { break }

  #yay, redir!
  set path $a
  if {($schm ne $s) || ($host ne $h) || ($port ne $p)} {
   set schm $s
   set rTLS $r
   set host $h
   if {[set ipv6 [expr {[string index $host 0] eq "\x5b"}]]} {
    regexp $static::DataDome_ipv6match $host junk haddr
   }
   set port $p
   set server "${host}:${port}"
   set hostname ""

   if {$proxy ne ""} {
    #one benefit of proxy is single persistent connection
    continue
   }
   if {$virt eq ""} {
    set dest ""
   }
  }
 } ; #rtry loop
 if {$rtry == -1} {
  set e "too many redirects for -uri ${uri}"
 }
 if {($e eq "") && (($rslt == 401) || ($rslt == 407))} {
  set e "missing/unsupported credentials for (WWW|Proxy)-Authenticate demand"
 }

 if {$state ne ""} {
  set stvar [list $connected $conn $dest \
                  $virt $proxy $nsvr $crumbs $creds $kopf]
 } elseif {$connected} {
  catch {close $conn}
  set connected false
 }

 #NOTE:  as of August 2013 f5 has not corrected an awkwardness affecting
 #all data obtained from SIDEBAND recv, which involves string content
 #conversions between Latin-1 and UTF-8.  The code in this proc *assumes*
 #the presence of that glitch and will work around it, but when f5 does
 #fix the issue we may have to revisit this code.
 if {($rbody ne "") || ($rformat ne "") || ($rheaders ne "")} {
  if {$sniff > 2} {
   incr sniff -3
   set charset [lindex [list "iso-8859-1" "utf-8" "binary"] $sniff]
  }
  while {$charset eq "unk"} {
   if {$mimetype eq "application/json"} {
    set charset "utf-8"
    break
   }

   set charset "binary" ; #fallback, plus ensures loop-exit

   set take "a[expr {$bootylen & 0x3ff}]"
   set odor [binary format $take $booty]
   binary scan $odor c* miasma ; #scan *before* regexp or someone mangles

   if {$odor starts_with "\xef\xbb\xbf"} {
    set charset "utf-8"
    break
   }
   if {($odor starts_with "\xfe\xff") || ($odor starts_with "\xff\xfe")} {
    set charset "utf-16" ; #we neglect UTF-32
    break
   }

   if {[regexp $static::DataDome_quiver $odor contag] &&
       [regexp {(?i)[\x20\t]content=['\x22]?([^'\x22>]+)} $contag junk contype]} {
    set contype [string tolower $contype]
    regexp {^[^/;\x20\t]+/[^;\x20\t]+} $contype mymtype
    if {$mimetype eq "unk/unk"} { set mimetype $mymtype }

    if {![regexp {charset=([^;\x20\t]+)} $contype junk charset] &&
        (($mimetype starts_with "text/") ||
         ($mimetype eq "application/javascript") ||
         ($mimetype eq "application/ecmascript")) && ($sniff != 2)} {
     set charset [expr {($sniff == 0) ? "iso-8859-1" : "utf-8"}]
    }
    if {$charset ne "binary"} { break }
   }

   if {[regexp {/((xml)|([^;\x20\t]+[+]xml))} $mimetype]} {
    set charset "utf-8" ; #text/xml dfl is ASCII; UTF-8 common, incls. both
    #but next we check for overrides...
   }
   if {[regexp $static::DataDome_csetmeta $odor junk charset] ||
       [regexp $static::DataDome_xmlenc $odor junk charset]} {
    set charset [string tolower $charset]
   }
   if {$charset ne "binary"} { break }

   set may_ascii true
   set may_utf8 true
   set may_latin true
   set p 0
   foreach c $miasma {
    set c [expr {$c & 0xff}]
    if {$c > 0x7f} {
     set may_ascii false
     if {($c < 0xc0) && (($p < 0xc2) || ($p > 0xf4))} {
      set may_utf8 false
      if {!$may_latin} { break }
     }
     if {($c == 0x81) || ($c == 0x8d) ||
         ($c == 0x8f) || ($c == 0x90) || ($c == 0x9d)} {
      set may_latin false
      if {!$may_utf8} { break }
     }
    }
    set p $c
   }
   #no time/space here to sniff for, say, Big5
   if {$mimetype starts_with "text/"} {
    if {$may_utf8} {
     set charset "utf-8"
    } else {
     if {$may_latin} { set charset "iso-8859-1" }
    }
   } elseif {$may_ascii} {
    set charset "utf-8"
   }
  }
  if {($charset eq "windows-1252") ||
      ([string first "ascii" $charset] >= 0)} {
   set charset "iso-8859-1"
  }

  if {$rformat ne ""} {
   upvar 1 $rformat rtn_fmt

   set rtn_fmt [expr {(($charset eq "utf-8") || ($charset eq "iso-8859-1")) ?
                      "text" : "binary"}]
  }
  if {$rbody ne ""} {
   upvar 1 $rbody rtn_body

   if {$charset eq "utf-8"} {
    set rtn_body [URI::decode [URI::encode $booty]] ; #don't a-ask me why!
   } elseif {$charset eq "iso-8859-1"} {
    #What is it?
    #Evil, pure and simple, from the Eighth Dimension!
    set rtn_body ""
    append rtn_body $booty
    ###regexp {.*} $booty rtn_body
   } else {
    set rtn_body [binary format a* $booty]
   }
  }
 }
 if {$rheaders ne ""} {
  upvar 1 $rheaders rtn_headers

  lappend hlist "X-DataDome-Charset" $charset
  set rtn_headers $hlist
 }
 if {$rcookies ne ""} {
  upvar 1 $rcookies rtn_cookies

  set cookies [list]
  set now [clock clicks -milliseconds]
  foreach {n c} $crumbs {
   if {[lindex $c 4] > $now} {
    lappend cookies $n [lindex $c 3]
   }
  }
  set rtn_cookies $cookies
 }

 if {$e ne ""} {
   call datadome_err_log "${tkm} ${e}"
 }
 return $rslt
 #At this point, http_req return value is HTTP status and
 #you may inspect your -(rbody|rheaders|rcookies|rformat)
 #variables for the results of your request
} ; #http_req


# After final [call /Common/HSSR::http_req -state xyz ...]
# SIDEBAND connection to remote server may remain open.
# Close it thusly:  [call /Common/HSSR::http_close xyz]
# (For lagniappe we also support [call /Common/HSSR::http_close $xyz])
#
proc http_close {stvar} {
 if {[catch {llength $stvar} varlen] || ($varlen == 1)} {
  upvar 1 $stvar state
 } else {
  set state $stvar
 }

 if {[catch {llength $state} stlen] || ($stlen < 9)} {
  set e "http_close argument must be xyz (or \x24xyz) from: http_req -state xyz"
  call datadome_err_log $e
  return -code error $e
 }
 if {[lindex $state 0]} {
  catch {close [lindex $state 1]}
 }
 set state [list false {} "" "" "" "" [list] "" ""]
 return
} ; #http_close

#== End of Mark's HTTP Super SIDEBAND Requestor! (v2.1) ==============
